//
//
//  fibonacci.cc
//
//      AUTHOR: Russell Lowke   Mon Feb 21st 2004
//
// Description: Calculates the Fibonacci sequence
//
//


#include <iostream>
#include <ctime>
#include <vector>
#include <cmath>

using namespace std;


typedef vector<int>    IntList;
typedef vector<IntList> Matrix;
typedef vector<Matrix> MatrixList;

enum {OVERFLOW_AT2_16 = 65536};
enum {OVERFLOW_AT2_32 = 4294967296};	// not used, but nice to know

enum FibType{RECURSIVE, ITERATIVE, MATRIX};

//
//  recursive call that calculates the fibonacci sequence
//
unsigned int fib_r (int n) {
    
    if (n <= 0)					// base case 0       = 0
		return 0;
	else if (n <= 2)			// base case 1 or 2  = 1
		return 1;
	else
		return ( (fib_r(n - 1) + fib_r(n - 2)) % OVERFLOW_AT2_16 );	// recursion
}


//
//  iterative call that calculates the fibonacci sequence
//
unsigned int fib_i (int n) {

    IntList f(n + 1);			// +1 as [0] included
    
    f[0] = 0;
    f[1] = 1;
    
	for (int i = 2; i <= n; ++i)
        f[i] = ( f[i - 1]  + f[i - 2] );
    
    return f[n];		// OVERFLOW_AT2_16 avoids overflow
}


//
//  utility to create a matrix of specified width and height from an array
//
Matrix buildMatrix(int width, int height, int data[]) {
    
    Matrix n;
    
    for (int x = 0; x < width; ++x) {
        IntList col;
        for (int y = 0; y < height; ++y)
            col.push_back( data[ width * y + x] );
        n.push_back( col );
    }
    
    return n;
}    

//
//  utility to print a Matrix
//
void printMatrix(Matrix lastTwo) {

    int width  = lastTwo.size();
    int height = lastTwo[0].size();
    
    for (int y = 0; y < height; ++y) {
        cout << "[ ";
        
        for (int x = 0; x < width; ++x) {
            cout << lastTwo[x][y] << " ";
        }
        
        cout << "]" << endl;
    }
    
    cout << endl;
}


//
// multiply matrices   m x n
//
Matrix muliplyMatrices (Matrix matrixM, Matrix matrixN) {

    Matrix n;
    
    // resultant matrix will have m's height and n's width
    int width  = matrixN.size();
    int height = matrixM[0].size();
    
    for (int x = 0; x < width; ++x) {
    
        IntList col;
        
        for (int y = 0; y < height; ++y ) {
        
            int val = 0;
            for (int i = 0; i < height; ++i )
                val += (matrixM[i][y] % OVERFLOW_AT2_16 * matrixN[x][i] % OVERFLOW_AT2_16 );
            
            // % OVERFLOW_AT2_16 avoids overflow
            
            col.push_back( val );
        }
        
        n.push_back( col );
    }
    
    return n;
}


//
//  matrix call that calculates a fibonacci number
//
int fib_m (int n) {
    
    if (n == 0) {
        return 0;
    } else {
        //  matrixA = [ 0 1 ]
        //            [ 1 1 ]
        //
        int dataA[] = {0, 1, 1, 1};
        Matrix matrixA = buildMatrix(2, 2, dataA );
        Matrix matrixB = matrixA;
        
        /*
        // matrixA^n      simple version for iterative matrix (not used)
        for (int i = 1; i < n; ++i )
            matrixA = muliplyMatrices(matrixA, matrixB);
        */
        
        MatrixList mList;
        matrixA = buildMatrix(2, 2, dataA );
        matrixB = matrixA;
        
        mList.push_back( matrixA );         // mList[0]
        int place = 0;
        int indice = 1;						// as in x^1
        
        // use repeated squaring
        do {
            indice *= 2;					// double the indice
            ++place;
            matrixA = muliplyMatrices(matrixA, matrixB);
            mList.push_back( matrixA );
            matrixB = matrixA;
        } while (indice*2 < n);
        
        for (; place >= 0; place--) {
        
            // get the indice of this "place"     2^place
            int pwr = (int) pow(2.0, place);
            
            if (indice + pwr <= n) {		// include if < n
                matrixA = muliplyMatrices(matrixA, mList.at(place));
                indice += pwr;				// update indice
            }
        }
        
        //  matrixC = [ 0 ]
        //            [ 1 ]
        //
        int dataC[] = {0, 1};
        Matrix matrixC = buildMatrix(1, 2, dataC);
        
        return muliplyMatrices(matrixA, matrixC)[0][0];
    }
}


void wait ( int seconds ) {
	clock_t endwait;
	endwait = clock() + seconds * CLK_TCK ;
	while (clock() < endwait) {}
}

//
// Calculates the number of fibonacci numbers calculated in [seconds] seconds
//
int nFibsInTime(int seconds) {

    int n = 0;
	clock_t endwait =  clock() + seconds * CLK_TCK;
    while (clock() < endwait) {
        ++n;
        fib_m(n);
    }
    
    return n;

}

//
// Calculates to the [fibNum] fibonacci number [nTimes] times
//  returning the duration taken
//
clock_t timedFib(int fibNum, int nTimes) {
    clock_t startTime = clock();
    
    for (int i; i < nTimes; i++ )
        fib_m(fibNum);
    clock_t endTime = clock();
    
    return (endTime - startTime);
}

void timedFib( int seconds, FibType type ) {
    clock_t endwait =  clock() + seconds * CLK_TCK;
    
    int n = 0;
    int r;
    
    while (clock() < endwait) {
        ++n;
        
        switch (type) {
            case RECURSIVE:
                r = fib_r(n);
                break;
            case ITERATIVE:
                r = fib_i(n);
                break;
            case MATRIX:
                r = fib_m(n);
                break;
        }
    }
    
    cout << "After " << seconds << " seconds fib_" << type << " computed  " << n 
         << " fibonacci sequences." << endl;
    cout << n << ") = " << r << endl; 
}

int main() {

    // short is 2^16
    // int and long, on the machine I'm working on, are 2^32
    
	// 16bit MAXINTEGER [short] = 2^16 = 65536        - 1 [for zero] = 65535
    // 32bit MAXINTEGER         = 2^32 = 4294967296   - 1 [for zero] = 4294967295
    // cout << INT_MAX from <climits>  = 2147483647  which is signed.
    
	// n = 24) 46368        is the last fibonacci number before unsigned short integer overflow [2^16 on this machine].
    //						the 25th number is 75025, which is > 65536
    // n = 47) 2971215073   is the last fibonacci number before unsigned integer overflow [2^32 on this machine].

    
    for (int n = 0; n < 30; n++) {
        // recursive
        cout << n <<"th fibonacci # = " << fib_r(n) << " [recursive]" << endl;
    }
    
    for (int n = 0; n < 30; n++) {
        // iterative
        cout << n <<"th fibonacci # = " << fib_i(n) << " [iterative]" << endl;
    }
    
	for (int n = 0; n < 30; n++) {
        // matrix
        cout << n <<"th fibonacci # = " << fib_m(n) << " [matrix]" << endl;
    }
    
    return 0;
}


    /* [ using iterator ]
	// for (IntList::iterator i = fib_vals.begin(); i != fib_vals.end(); ++i)
	//	 cout << "iterative #? ) " << *i << endl;	// print
    */
    