//
//
//  fibonacci_m.cc
//
//      AUTHOR: Russell Lowke   Mon Feb 21st 2004
//
// Description: Calculates the Fibonacci sequence using matrix
//
//

#include <iostream>
#include <ctime>
#include <vector>
#include <cmath>

using namespace std;

typedef vector<int>    IntList;
typedef vector<IntList> Matrix;
typedef vector<Matrix> MatrixList;

enum {OVERFLOW_AT2_16 = 65536};
enum {OVERFLOW_AT2_32 = 4294967296};	// not used, but nice to know

//
//  utility to create a matrix of specified width and height from an array
//
Matrix buildMatrix(int width, int height, int data[]) {
    
    Matrix n;
    
    for (int x = 0; x < width; ++x) {
        IntList col;
        for (int y = 0; y < height; ++y)
            col.push_back( data[ width * y + x] );
        n.push_back( col );
    }
    
    return n;
}

//
// multiply matrices   m x n
//
Matrix muliplyMatrices (Matrix matrixM, Matrix matrixN) {

    Matrix n;
    
    // resultant matrix will have m's height and n's width
    int width  = matrixN.size();
    int height = matrixM[0].size();
    
    for (int x = 0; x < width; ++x) {
    
        IntList col;
        
        for (int y = 0; y < height; ++y ) {
        
            int val = 0;
            for (int i = 0; i < height; ++i )
                val += (matrixM[i][y] % OVERFLOW_AT2_16 * matrixN[x][i] % OVERFLOW_AT2_16 );
            
            // % OVERFLOW_AT2_16 avoids overflow
            
            col.push_back( val );
        }
        
        n.push_back( col );
    }
    
    return n;
}

//
//  matrix call that calculates a fibonacci number
//
int fib_m (int n) {
    
    if (n == 0) {
        return 0;
    } else {
        //  matrixA = [ 0 1 ]
        //            [ 1 1 ]
        //
        int dataA[] = {0, 1, 1, 1};
        Matrix matrixA = buildMatrix(2, 2, dataA );
        Matrix matrixB = matrixA;
        
        /*
        // matrixA^n      simple version for iterative matrix (not used)
        for (int i = 1; i < n; ++i )
            matrixA = muliplyMatrices(matrixA, matrixB);
        */
        
        MatrixList mList;
        matrixA = buildMatrix(2, 2, dataA );
        matrixB = matrixA;
        
        mList.push_back( matrixA );         // mList[0]
        int place = 0;
        int indice = 1;						// as in x^1
        
        // use repeated squaring
        do {
            indice *= 2;					// double the indice
            ++place;
            matrixA = muliplyMatrices(matrixA, matrixB);
            mList.push_back( matrixA );
            matrixB = matrixA;
        } while (indice*2 < n);
        
        for (; place >= 0; place--) {
        
            // get the indice of this "place"     2^place
            int pwr = (int) pow(2.0, place);
            
            if (indice + pwr <= n) {		// include if < n
                matrixA = muliplyMatrices(matrixA, mList.at(place));
                indice += pwr;				// update indice
            }
        }
        
        //  matrixC = [ 0 ]
        //            [ 1 ]
        //
        int dataC[] = {0, 1};
        Matrix matrixC = buildMatrix(1, 2, dataC);
        
        return muliplyMatrices(matrixA, matrixC)[0][0];
    }
}

//
// Calculates the number of fibonacci numbers calculated in [seconds] seconds
//
int nFibsInTime(int seconds) {

    int n = 0;
	clock_t endwait =  clock() + seconds * CLK_TCK;
    while (clock() < endwait) {
        ++n;
        fib_m(n);
    }
    
    return n;

}

//
// Calculates to the [fibNum] fibonacci number [nTimes] times
//  returning the duration taken
//
clock_t timedFib(int fibNum, int nTimes) {
    clock_t startTime = clock();
    
    for (int i; i < nTimes; i++ )
        fib_m(fibNum);
    clock_t endTime = clock();
    
    return (endTime - startTime);
}

int main() {
    
    clock_t duration = timedFib(24, 10000);
    cout << "It took " << duration << " 1/" << CLK_TCK <<
        "ths of a second for fib_m to compute the 24th fibonacci number 10000 times" << endl;
    
    cout << endl;
    
    int n = nFibsInTime(60);
	cout << "After 1 minute fib_m computed " << n << " fibonacci numbers." << endl;
    
    return 0;
}
    