///////////
//
//  Graph.cpp
//
//      AUTHOR: Russell James Lowke  (lowke@fas.harvard.edu)
//
//		  Date: 3/13/2004
//

#include "graph.hpp"								// include header
#include "edge.hpp"
#include "vertex.hpp"

#include <iostream>
#include <cmath>

using namespace std;
double randomDble();
double randomInt();


//
//  constructor for Graph class
//
Graph :: Graph() {}
Graph :: ~Graph() {
    for (VertexList::iterator iter = vertices.begin(); iter!= vertices.end(); ++iter) {
        int initialsize = vertices.size();
        if ( (*iter)->getName() <= initialsize ) {
            delete *iter;
        }
    }
}


void Graph :: addVertex(Vertex* mVertex) {
	vertices.push_back(mVertex);
}


void Graph :: addVertex(double mX, double mY) {
	static int vertexNum = 0;
	addVertex( new Vertex(vertexNum++, mX, mY));
}


void Graph :: addEdge(Edge mEdge) {
	edges.push_back(mEdge);
}


void Graph :: createEdge(int mStart, double mWeight, int mEnd) {
	addEdge(Edge(vertices[mStart], mWeight, vertices[mEnd]));
}

void Graph :: createEdge(Vertex* mStart, double mWeight, Vertex* mEnd) {
	addEdge(Edge(mStart, mWeight, mEnd));
}

void Graph :: printVertices() {
	cout << "Printing graph vertices:\n";
	for (VertexList::iterator i = vertices.begin(); i != vertices.end(); ++i)
		(*i)->pprint();
}


void Graph :: printEdges() {
	cout << "Printing graph edges:\n";
	for (EdgeList::iterator i = edges.begin(); i != edges.end(); ++i)
		i->pprint();
}


void Graph :: sortEdges() {
	//	Quicksort adapted from "C Programming - A Modern Approach" by K.N.King
	edgeQuicksort(0, edges.size() - 1);
}


int Graph :: edgeSplit(int low, int high) {
	Edge partElement = edges[low];
	
	for(;;) {
		while (low < high && partElement.getEdgeWeight() <= edges[high].getEdgeWeight())
			--high;
		if (low >= high) break;
		edges[low++] = edges[high];
		
		while (low < high && edges[low].getEdgeWeight() <= partElement.getEdgeWeight())
			++low;
		if (low >= high)
			break;
		edges[high--] = edges[low];
	}
	
	edges[high] = partElement;
	
	return high;
}


void Graph :: edgeQuicksort(int low, int high) {
	int middle;
	
	if (low >= high)
		return;
	middle = edgeSplit(low, high);
	edgeQuicksort(low, middle - 1);
	edgeQuicksort(middle + 1, high);
}

void Graph :: generateVertices(int n) {
	for (int i = 0; i < n; ++i)
		addVertex(randomDble(), randomDble());
}

// used in construction of unit square graphs
// connect every vertex with every other via an edge.
void Graph :: connectVertices(bool distWeight) {
	
	int vertexCount = vertices.size();
	
	double mWeight;
	Vertex* thisVertex;
	Vertex* otherVertex;
	
	for (int i = 0; i < vertexCount; i++) {
		thisVertex = vertices[i];
		
		for (int j = i+1; j < vertexCount; j++) {
			otherVertex = vertices[j];
			
			mWeight = distWeight ? distance(thisVertex, otherVertex) : randomDble();
			createEdge(thisVertex, mWeight, otherVertex);
		}
	}
}

// distance between thisVertex and thatVertex
double Graph :: distance(Vertex* thisVertex, Vertex* otherVertex) {
	return sqrt( pow(abs(otherVertex->getX() - thisVertex->getX()), 2.0) + 
				 pow(abs(otherVertex->getY() - thisVertex->getY()), 2.0) );
}

//
//  accessors
//
VertexList	Graph :: getVertices() const     { return vertices; }
EdgeList	Graph :: getEdges() const       { return edges; }