//
//
//  Heuristics.cpp
//
//      AUTHOR: Russell Lowke
//
//		  Date: 5/15/2004
//
//  Implementing some randomized and local search heuristics for solving NP-complete problems.
//


#include <iostream>
using std::cin;
using std::cout;
using std::endl;

#include <NTL/ZZ.h>
#include <NTL/RR.h>
using namespace NTL;
typedef RR bigdouble;
typedef ZZ bigint;

#include <queue>
#include <vector>
typedef std::priority_queue< bigint > PriorityQueue;

const int MAX_ITER = 250000;
const float MATH_E = 2.71828182846;

//
//  random generator
//
void randomizer()
{
    time_t      timeval;
    
    time(&timeval);
    srand(timeval);
    SetSeed(to_ZZ(rand()));
}

double randomReal() {
    return rand() / (double) RAND_MAX;
}

int randomInt(int min, int max) {
    int range = (max + 1) - min;
    return (int) (randomReal() * range + min);
}


/* produces an array of 100 randomly generated integers in the range [1, 10^12]*/
void generateRandomInstance(unsigned long randomInstance[100]) {

    long bound = (long) pow(10, 12);

    bigint bi = to_ZZ("1203471290348710298347102840189347012978340198734");
    cout << "The bigint in generateRandomInstance, bigbound, is: " << bi << endl;
    for (int i= 0; i < 100; ++i) {
	
        unsigned long current = (rand() % bound) + 1;
        /* for creating 64 random bits
        current << 32;
        current |= ((rand() % bound) + 1);
        */
        randomInstance[i] = current;
    }
}

/* produces an array of 100 randomly generated integers in the range [1, 10^12]*/
void generateRandomInstance(bigint randomInstance[100]) {

    //long bound = (long) pow(10, 12);
    bigint bound = power_ZZ(10, 12);

    for (int i= 0; i < 100; ++i) {
        bigint current = RandomBnd(bound) + 1;
        randomInstance[i] = current;
    }
    
    cout << "Instance is:\n{ ";
    for (int i = 0; i < 100; ++i){
        cout << randomInstance[i] << " ";
    }
    cout << "}" << endl;
    
}


void generateRandomInts(int randomInstance[100], int from, int to) {
	
	for (int i = 0; i < 100; ++i) {
		randomInstance[i] = randomInt(from, to);
	}
}


void copyArray(int arraySource[100], int arrayDest[100]) {
    for (int i = 0; i < 100; ++i) {
        arrayDest[i] = arraySource[i];
    }
}


void copyBigIntArray(bigint arraySource[100], bigint arrayDest[100]) {
    for (int i = 0; i < 100; ++i) {
        arrayDest[i] = arraySource[i];
    }
}


bigint* doGroupings(bigint instance[100], int groupings[100]) {
	
	bigint* result = new bigint[100];
	copyBigIntArray(instance, result);

	for (int i = 0; i < 100; ++i) {
		int firstCaseOf = -1;

		for (int j = 0; j < 100; ++j) {
			
			if (groupings[j] == i) {
				if (firstCaseOf == -1) {
					firstCaseOf = j;
				} else {
					result[firstCaseOf] += result[j];
					result[j] = 0;
				}
			}
		}
	}
	
	return result;
}


/* produces an array of 100 randomly generated integers which are either -1 or +1 */
void generateRandomSolution(int randomSolution[100]) {
    for (int i = 0; i < 100; ++i) {
        int num = rand() % 2;
        if (num == 0) {
			num = -1;
		}
        randomSolution[i] = num;
    }
    return;
}


/* must pass in a copy of original data.  It will be modified into a neighbor */
void generateRandomNeighbor(int newNeighbor[100]) {
    int i = random() % 100 + 1;
    int j;
    do {
        j = random() % 100 + 1;
    } while (i == j);
	
    //now that i and j are different indices, flip i
    newNeighbor [i] *= -1;
	
    //and flip j with 50% probability
    if ( randomReal() < 0.50) {
        newNeighbor[j] *= -1;
    }
}

void generateRandomNeighborFromP(int newNeighbor[100]) {
    int i = random() % 100 + 1;
    int j;
    for (int k = 0; k < 100; ++k) {
        j = random() % 100 + 1;
		
		if (newNeighbor[i] != newNeighbor[j]){
			newNeighbor[i] = newNeighbor[j];
			break;
		}
    }
}


long calculateResidue(unsigned long instance[100], int assignments[100]) {
    long residue = 0;
    for (int i = 0; i < 100; ++i) {
        residue += instance[i] * assignments[i];
    }
    return abs(residue);
}


bigint calculateResidue(bigint instance[100], int assignments[100]) {
    bigint residue;
    for (int i = 0; i < 100; ++i) {
        residue += instance[i] * assignments[i];
    }
    return abs(residue);
}


/* implementing the KK heuristic */
bigint KKsolution(bigint instance[100]){

	// make a priority queue of size 100;
	// feed all of the numbers from instance into it
	PriorityQueue queue;
	for(int i = 0; i < 100; ++i) {
		if ( instance[i] != 0 ){
			queue.push(instance[i]);
		}
	}
	
	bigint a, b, diff;

	while (queue.size() > 2) {
		a = queue.top();
		queue.pop();
		
		b = queue.top();
		queue.pop();
		if (b == 0) {
			queue.push(a);
			break;
		}

		diff = abs(a - b);
		queue.push(diff);
	}
	
	return (queue.top());
}


bigint calculateResidueWithP(bigint instance[100], int assignments[100]) {
    bigint residue;
	bigint* temp = doGroupings(instance, assignments);
    residue = KKsolution(temp);
	delete[] temp;
    return abs(residue);
}


int* randomSolution(unsigned long* instance) {

    // Start with a random solution S
    int* s1 = new int[100];
    int* s2;
    unsigned long r1, r2;
    generateRandomSolution(s1);
    r1 = calculateResidue(instance, s1);
    
    // for iter = 1 to max_iter
    //		S` = a random solution
    for (int i = 0; i < MAX_ITER; ++i) {
        s2 = new int[100];
        //
        generateRandomSolution(s2);
        r2 = calculateResidue(instance, s2);
        
        // if residue(S`) < residue(S) then S = S`
        if (r2 < r1) {
            delete[] s1;
            s1 = s2;
            r1 = r2;
        } else {
            delete[] s2;
        }
    }
    
    // resturn S
    return s1;	
}

int* randomSolutionWithP(bigint* instance) {

    // Start with a random solution S
    int* s1 = new int[100];
	generateRandomInts(s1, 1, 100);
    int* s2;
    bigint r1, r2;
    r1 = calculateResidueWithP(instance, s1);
    
    // for iter = 1 to max_iter
    //		S` = a random solution
    for (int i = 0; i < MAX_ITER; ++i) {
        s2 = new int[100];
        
        generateRandomSolution(s2);
		generateRandomInts(s2, 1, 100);
        r2 = calculateResidueWithP(instance, s2);
        // if residue(S`) < residue(S) then S = S`
        if (r2 < r1) {
            delete[] s1;
            s1 = s2;
            r1 = r2;
			
        } else {
            delete[] s2;
        }
    }
    
    // resturn S
    return s1;	
}

int* randomSolution(bigint* instance) {

    // Start with a random solution S
    int* s1 = new int[100];
    int* s2;
    bigint r1, r2;
    generateRandomSolution(s1);
    r1 = calculateResidue(instance, s1);
    
    // for iter = 1 to max_iter
    //		S` = a random solution
    for (int i = 0; i < MAX_ITER; ++i) {
        s2 = new int[100];
        
        generateRandomSolution(s2);
        r2 = calculateResidue(instance, s2);
        
        // if residue(S`) < residue(S) then S = S`
        if (r2 < r1) {
            delete[] s1;
            s1 = s2;
            r1 = r2;
        } else {
            delete[] s2;
        }
    }
    
    // resturn S
    return s1;	
}

int* hillClimbing(bigint* instance) {

    // Start with a random solution S
    int* s1 = new int[100];
    int* s2;
    bigint r1, r2;
    generateRandomSolution(s1);
    r1 = calculateResidue(instance, s1);
	
    // for iter = 1 to max_iter
    for (int i = 0; i < MAX_ITER; ++i) {

        //		S` = a random neighbor of S
        s2 = new int[100];
        copyArray(s1, s2);
        generateRandomNeighbor(s2);
        r2 = calculateResidue(instance, s2);

        // if residue(S`) < residue(S) then S = S`
        if (r2 < r1) {
			//cout<<"disposing of s1, assigning s2 to s1"<<endl;
            delete[] s1;
            s1 = s2;
            r1 = r2;
        } else {
			//cout<<"disposing of s2"<<endl;
            delete[] s2;
        }
    }

    // resturn S
    return s1;
}

int* hillClimbingWithP(bigint* instance) {

    // Start with a random solution S
    int* s1 = new int[100];
    int* s2;
    bigint r1, r2;
	generateRandomInts(s1, 1, 100);
    r1 = calculateResidueWithP(instance, s1);

    // for iter = 1 to max_iter
    for (int i = 0; i < MAX_ITER; ++i) {

        //		S` = a random neighbor of S
        s2 = new int[100];
        copyArray(s1, s2);
        generateRandomNeighborFromP(s2);
        r2 = calculateResidueWithP(instance, s2);

        // if residue(S`) < residue(S) then S = S`
        if (r2 < r1) {

			if (s1 != s2){
				delete[] s1;
			}
            s1 = s2;
            r1 = r2;
        } else {

            delete[] s2;
        }
    }

    // resturn S
    return s1;
}

float temperature(int iter){
    // the real temperature calculation
    return (pow(10, 10) * pow(0.9, (iter/300.0)));
}


int* simulatedAnnealing(bigint* instance) {

    // Start with a random solution S
    int* s1 = new int[100];
    int* s2;
    bigint r1, r2, r3;
    generateRandomSolution(s1);
    r1 = calculateResidue(instance, s1);
	float temper;
    // S`` = S
    int* s3 = new int[100];
    copyArray(s1, s3);
    r3 = r1;

    // for iter = 1 to max_iter
    for (int i = 0; i < MAX_ITER; ++i) {

        //		S` = a random neighbor of S
        s2 = new int[100];
        copyArray(s1, s2);
        generateRandomNeighbor(s2);
        r2 = calculateResidue(instance, s2);

        // if residue(S`) < residue(S) then S = S`
        if (r2 < r1) {
			if (s1 != s3) {
				delete[] s1;
			}
            s1 = s2;
			r1 = r2;
        } else {
		
            // S = S` with probability exp(-(res(S`) - res(S`)) / T(iter))
			temper = (temperature(i) > 50 ? temperature(i) : 50);
			
            if (randomReal() < power(to_RR(MATH_E), ( - (to_long(r2 - r1)) / temper ))) {
				//			power (float	,   zz / float
				if (s1 != s3) {
					delete[] s1;
				}
                s1 = s2;
				r1 = r2;
            } else {
                delete[] s2;
            }
        }

        // if residue(S) < residue(S``) then S = S`
        if (r1 < r3) {
            delete[] s3;
            s3 = s1;
            r3 = r1;
        }
    }
    
    if (s1 != s3) {
        delete[] s1;
    }
    
    // resturn S``
    return s3;
}

int* simulatedAnnealingWithP(bigint* instance) {

    // Start with a random solution S
    int* s1 = new int[100];
    int* s2;
	float temper;
    bigint r1, r2, r3;
	generateRandomInts(s1, 1, 100);
    r1 = calculateResidueWithP(instance, s1);

    // S`` = S
    int* s3 = new int[100];
    copyArray(s1, s3);
    r3 = r1;

    // for iter = 1 to max_iter
    for (int i = 0; i < MAX_ITER; ++i) {

        //		S` = a random neighbor of S
        s2 = new int[100];
        copyArray(s1, s2);
        generateRandomNeighborFromP(s2);
        r2 = calculateResidueWithP(instance, s2);

        // if residue(S`) < residue(S) then S = S`
        if (r2 < r1) {
			if (s1 != s3) {
				delete[] s1;
			}
            s1 = s2;
			r1 = r2;
        } else {
		
            // S = S` with probability exp(-(res(S`) - res(S`)) / T(iter))
			temper = (temperature(i) > 50 ? temperature(i) : 50);
			
            if (randomReal() < power(to_RR(MATH_E), ( - (to_long(r2 - r1)) / temper ))) {
				//			power (float	,   zz / float
				if (s1 != s3) {
					delete[] s1;
				}
                s1 = s2;
				r1 = r2;
            } else {
                delete[] s2;
            }
        }

        // if residue(S) < residue(S``) then S = S`
        if (r1 < r3) {
            delete[] s3;
            s3 = s1;
            r3 = r1;
        }
    }
    
    if (s1 != s3) {
        delete[] s1;
    }
    
    // resturn S``
    return s3;
}
